import { MiddlewareConsumer, Module, NestModule, ParseUUIDPipe } from '@nestjs/common';
import { ConfigModule } from '@nestjs/config';
import { APP_FILTER } from '@nestjs/core';
import { TypeOrmModule } from '@nestjs/typeorm';
import * as cookieSession from 'cookie-session';
import { NextFunction, Request, Response } from 'express';
import { Connection, ConnectionOptions, getConnectionOptions } from 'typeorm';

import { AppController } from './app.controller';
import { AppService } from './app.service';
import { ChatbotModule } from './chatbot/chatbot.module';
import { EventsModule } from './events/events.module';
import { HttpExceptionFilter } from './http-exception.filter';
import { JobsModule } from './jobs/jobs.module';
import { SessionService } from './user/session.service';
import { UserModule } from './user/user.module';

@Module({
  imports: [
    ConfigModule.forRoot(),
    TypeOrmModule.forRootAsync({
      useFactory: async () => {
        const opts = await getConnectionOptions();

        /**
         * Make sure that the production build uses the js files.
         */
        if (__filename.endsWith('js')) {
          return {
            ...opts,
            entities: ['build/**/*.entity.js'],
            migrations: ['build/migration/**/*.js'],
            subscribers: ['build/subscriber/**/*.js']
          } as ConnectionOptions;
        }
        return opts;
      }
    }),
    JobsModule,
    ChatbotModule,
    EventsModule,
    UserModule
  ],
  controllers: [AppController],
  providers: [
    {
      provide: APP_FILTER,
      useClass: HttpExceptionFilter
    },
    ParseUUIDPipe,
    AppService
  ]
})
export class AppModule implements NestModule {

  constructor(
    private readonly connection: Connection,
    private sessionService: SessionService) {
    this.connection.showMigrations().then((hasMigration) => {
      if (hasMigration === true) {
        console.warn('The database schema is not up to date.');
        console.warn('Please run `npm run orm -- migration:run` to update it.');
      }
    });
  }

  public configure(consumer: MiddlewareConsumer) {
    consumer
      .apply(
        /**
         * Cookie middleware for reading and writing cookies.
         * @TODO remove httpOnly
         * @TODO .env based keys
         */
        cookieSession({
          name: 'session',
          keys: ['test'],
          maxAge: 365 * 24 * 60 * 60 * 1000 // 365 days
        }),

        /**
         * Custom middleware for automatically fetching the session
         * of the currently authenticated user.
         */
        async (req: Request, res: Response, next: NextFunction) => {
          if (req.session?.id) {
            try {
              const session = await this.sessionService.one(req.session.id);
              req.userSession = session;
            } catch (error) { }
          }
          next();
        }).forRoutes('*');
  }

}
