import { Injectable } from '@nestjs/common';
import { ConfigService } from '@nestjs/config';
import { ModuleRef } from '@nestjs/core';
import { ActivityHandler, ActivityTypes, ChannelAccount, StatePropertyAccessor, TurnContext } from 'botbuilder';
import { DialogContext, DialogSet, DialogTurnResult, DialogTurnStatus, TextPrompt } from 'botbuilder-dialogs';

import { EventsService } from '../events/events.service';
import { SessionService } from '../user/session.service';
import { UserService } from '../user/user.service';

import { ChatbotService } from './chatbot.service';
import { EventDialog, HelpDialog, StartDialog } from './dialogs';
import { UserSettings } from './user';

@Injectable()
export class ChatbotHandler extends ActivityHandler {

    protected userSettings: StatePropertyAccessor<UserSettings>;
    protected dialogSet: DialogSet;

    constructor(
        private readonly botService: ChatbotService,
        private readonly configService: ConfigService,
        private readonly eventsService: EventsService,
        private readonly sessionService: SessionService,
        private readonly userService: UserService,
        private readonly moduleRef: ModuleRef) {
        super();

        this.userSettings = this.botService.userState.createProperty('UserSettings');
        this.dialogSet = new DialogSet(this.botService.conversationState.createProperty('DialogState'))
        .add(new HelpDialog(ChatbotHandler.name + HelpDialog.name))
        .add(new TextPrompt(ChatbotHandler.name + TextPrompt.name))
        .add(new StartDialog(ChatbotHandler.name + StartDialog.name, this.userSettings, this.userService, this.sessionService))
        .add(new EventDialog(ChatbotHandler.name + EventDialog.name, this.userSettings, this.eventsService, this.configService));

        // See https://aka.ms/about-bot-activity-message to learn more about the message and other activity types.
        this.onMessage(async (context, next) => {
            const dialogContext = await this.dialogSet.createContext(context);

            const userId = context.activity.from.id;
            if (!this.botService.conversations.has(userId)) {
                if (context.activity.from.name !== 'eventingHDM') {
                    this.botService.addConversationReference(context, { id: userId } as ChannelAccount);
                }
            }

            const results = await dialogContext.continueDialog();
            const hasNextCmd = context.activity.text.length && results.status === DialogTurnStatus.complete;

            if (results.status === DialogTurnStatus.empty || hasNextCmd) {
                await this.processSelection(dialogContext);
            }

            // By calling next() you ensure that the next BotHandler is run.
            await next();
        });

        this.onDialog(async (context, next) => {
            /* Save any state changes. The load happened during the execution of the Dialog. */
            await this.botService.conversationState.saveChanges(context, false);
            await this.botService.userState.saveChanges(context, false);

            /* By calling next() you ensure that the next BotHandler is run. */
            await next();
        });

        this.onMembersAdded(async (context, next) => {
            const membersAdded = context.activity.membersAdded;
            for (const member of membersAdded) {
                if (member.id !== context.activity.recipient.id) {
                    this.botService.addConversationReference(context, member);
                    context.activity.text = StartDialog.name;

                    const dialogContext = await this.dialogSet.createContext(context);
                    await this.processSelection(dialogContext);
                }
            }
            // By calling next() you ensure that the next BotHandler is run.
            await next();
        });

        this.eventsService.onReminder(async (eventId, userId) => {
            const user = await this.userService.one({ id: userId });
            const attendance = await this.eventsService.getAttendance(eventId, userId);
            await this.botService.continueConversation(user.telegramId, async (context) => {
                context.activity.text = EventDialog.name;
                context.activity.value = [{
                    ...attendance.event,
                    rsvp: attendance,
                    broadcast: 'Du hast mich gebeten dir an folgender Veranstaltung zu erinnern.'
                } as EventDialog.EventRecap];

                const dialogContext = await this.dialogSet.createContext(context);
                await this.processSelection(dialogContext);
            });
        });

        this.eventsService.onAttendee(async (attendance) => {
            this.botService.continueConversation(attendance.user.telegramId, async (context) => {
                context.activity.text = EventDialog.name;

                const statusString: EventDialog.ParticipationStrings = {
                        accepted: 'angemeldet',
                        maybe: 'mit Vorbehalt angemeldet',
                        declined: 'abgemeldet'
                    };

                context.activity.value = [{
                        ...attendance.event,
                        rsvp: attendance,
                        broadcast: 'Du hast dich gerade eben für folgende Veranstaltung ' + statusString[attendance.status] + '.'
                    } as EventDialog.EventRecap];

                const dialogContext = await this.dialogSet.createContext(context);
                await this.processSelection(dialogContext);
            });
        });

        /**
         * Broadcast to all users when new events get added.
         * TODO: implement organiser subscribing
         */
        this.eventsService.onEvent(async (event) => {
            this.botService.conversations.forEach(async (reference, key) => {
                await this.botService.continueConversation(key, async (context) => {
                    context.activity.text = EventDialog.name;
                    context.activity.value = [{
                        ...event,
                        broadcast: 'Folgende Veranstaltung wurde eben veröffentlicht, möchtest du daran teilnehmen?'
                    } as EventDialog.EventRecap];

                    const dialogContext = await this.dialogSet.createContext(context);
                    await this.processSelection(dialogContext);
                });
            });
        });
    }

    private async processSelection(dialogContext: DialogContext) {
        await dialogContext.context.sendActivity({ type: ActivityTypes.Typing });

        // get action
        let action: string = dialogContext.context.activity.text;
        const actionValue = action.split(',');
        let options;

        if (actionValue.length > 1) {
            action = actionValue.shift();
            options = actionValue;
        } else {
            action = action.replace('/', '');
            options = dialogContext.context.activity.value;
        }

        // enable custom dialog calling
        let turnResult: DialogTurnResult;
        switch (action) {
            case StartDialog.name:
            case 'start':
            case 'login':
            const member = dialogContext.context.activity.from;
            turnResult = await dialogContext.beginDialog(ChatbotHandler.name + StartDialog.name, { member });
            break;

            case 'cal':
            const user = await this.userSettings.get(dialogContext.context);
            await dialogContext.context.sendActivity('Deine Kalender URL lautet: ' + this.configService.get('HOME_URL') + '/calendars/' + user.publicId);
            break;

            case 'event':
                const event = await this.eventsService.repository.createQueryBuilder('event').orderBy('RAND()').limit(1).getOne();
                options = [ event ] as EventDialog.EventDialogOptions;
            case EventDialog.name:
            turnResult = await dialogContext.beginDialog(ChatbotHandler.name + EventDialog.name, options);
            break;

            default:
            return await dialogContext.beginDialog(
                ChatbotHandler.name + HelpDialog.name,
                { text: 'Das habe ich leider nicht verstanden.\
                Soll ich dir meine Liste von Befehlen anzeigen lassen?' } as HelpDialog.DialogOptions);
        }

        // enable overriding default feedback dialog
        switch (action) {
            case EventDialog.name:
            // don't display the feedback dialog
            return turnResult;
            break;

            default:
            return await dialogContext.beginDialog(ChatbotHandler.name + HelpDialog.name);
        }
    }
}
