import { ConfigService } from '@nestjs/config';
import { CardFactory, MessageFactory, StatePropertyAccessor } from 'botbuilder';
import { Choice, ChoiceFactory, ChoicePrompt, ComponentDialog, DialogContext, TextPrompt, WaterfallDialog } from 'botbuilder-dialogs';
import { DeepPartial } from 'typeorm';

import { EventAttendance, EventAttendee } from '../../events/attendee.entity';
import { Event } from '../../events/event.entity';
import { EventsService } from '../../events/events.service';
import { UserSettings } from '../user';

export class EventDialog extends ComponentDialog<EventDialog.EventDialogOptions> {

    constructor(
        readonly id: string,
        private readonly userSettings: StatePropertyAccessor<UserSettings>,
        private readonly eventsService: EventsService,
        private readonly configService: ConfigService) {
        super(id);

        // RSVP Dialogs
        this.addDialog(
            new WaterfallDialog<EventDialog.EventRecap>(id + EventDialog.EventDialogs.RSVP_DIALOG, [
                async (step) => {
                    const participations: EventDialog.ParticipationStrings = {
                        accepted: 'Akzeptieren',
                        maybe: 'Mit Vorbehalt',
                        declined: 'Ablehnen'
                    };

                    if ('rsvp' in step.options) {
                        delete participations[step.options.rsvp.status]; // remove the selected state
                    }

                    const choices = ChoiceFactory.toChoices(Object.values(participations).map((title, i) => {
                        const key = Object.keys(participations)[i];
                        return { value: key, action: { title, type: 'postBack', value: key }};
                    }));

                    return await step.prompt(this.id + 'cp', {
                        choices,
                        prompt: '**RSVP:**\n\n' + step.options.title,
                        retryPrompt: 'Das habe ich leider nicht verstanden'
                    });
                },
                async (step) => {
                    step.options.rsvp = await this.eventsService.setAttendance(
                        step.options.id,
                        step.options.rsvp.user.id,
                        step.result.value,
                        true);

                    const participations: EventDialog.ParticipationStrings = {
                        accepted: 'Zusage',
                        maybe: 'Zusage mit Vorbehalt',
                        declined: 'Absage'
                    };

                    await step.context.sendActivity(MessageFactory.text('Danke, deine '
                    + participations[step.result.value]
                    + ' wurde an dem Veranstalter übermittelt'));
                    // await this.sendEventCard(step.parent, step.options);
                    // prevent any previous user input from getting interpreted as a command
                    step.context.activity.text = '';
                    return step.endDialog();
                }
            ])
        );

        const cancelChoice: Choice = { value: 'cancel', action: { title: 'Abbrechen', type: 'postBack', value: 'cancel' }};
        const deleteChoice: Choice = { value: 'delete', action: { title: 'Erinnerung löschen', type: 'postBack', value: 'delete' }};

        // Reminder dialogs
        this.addDialog(
            new WaterfallDialog<EventDialog.EventRecap>(id + EventDialog.EventDialogs.SET_REMINDER_DIALOG, [
                /* async (step) => {
                    return await step.prompt(this.id + 'tp', 'In wie viele Minuten findet das Event statt?');
                }, */
                async (step) => {
                    const localTimestamp = (step.context.activity.localTimestamp || new Date()).valueOf();
                    // step.options.eventDate = localTimestamp + Number(step.result) * 60 * 1000;
                    let choices: Array<[string, number]> = [];
                    if (this.isInDateRange(step.options.dateStart, 60 * 24 * 10)) {
                        choices = [
                            ['Eine Woche davor', 60 * 24 * 7],
                            ['2 Tage davor', 60 * 24 * 2],
                            ['1 Tag davor', 60 * 24],
                            ['An dem Tag um 9 Uhr', 9]
                        ];
                    } else if (this.isInDateRange(step.options.dateStart, 60 * 24 * 3)) {
                        choices = [
                            ['1 Tag davor', 60 * 24],
                            ['An dem Tag um 9 Uhr', 9],
                            ['2 Std. davor', 60 * 2],
                            ['1 Std. davor', 60]
                        ];
                    } else if (this.isInDateRange(step.options.dateStart, 30)) {
                        choices = [
                            ['An dem Tag um 9 Uhr', 9],
                            ['6 Std. davor', 60 * 6],
                            ['2 Std. davor', 60 * 2],
                            ['1 Std. davor', 60],
                            ['30 Min. davor', 30]
                        ];
                    }

                    const promptChoices: Choice[] = [];
                    for (const choice of choices) {
                        const title = choice[0];
                        const dateDiff = choice[1]; // diff time in minutes
                        const toMilli = 60 * 1000;
                        const date = new Date(step.options.dateStart);

                        if (dateDiff <= 24 || dateDiff >= 60 * 24) {
                            date.setHours(dateDiff, 0, 0);
                        } else {
                            date.setTime(step.options.dateStart - dateDiff * toMilli);
                        }

                        if (dateDiff >= 60 * 24) {
                            date.setTime(date.valueOf() - dateDiff * toMilli);
                        }

                        if (date.valueOf() < (localTimestamp + 60 * 1000)) {
                            continue;
                        }

                        const key = date.toUTCString();
                        promptChoices.push({ value: key, action: { title, type: 'postBack', value: key }});
                    }

                    if (!promptChoices.length) {
                        const notice = 'Ich kann dich leider nicht mehr an dieser Veranstaltung erinnern.\
                        Entweder sie beginnt gleich oder ihr Beginn liegt bereits in der Vergangenheit';
                        await step.context.sendActivity(notice);
                        // prevent any previous user input from getting interpreted as a command
                        step.context.activity.text = '';
                        return await step.endDialog();
                    }

                    if (step.options?.reminder) {
                        promptChoices.push(deleteChoice);
                    }
                    promptChoices.push(cancelChoice);

                    return await step.prompt(this.id + 'cp', {
                        choices: promptChoices,
                        prompt: '**Erinnerung setzen:**\n\n' + step.options.title,
                        retryPrompt: 'Das habe ich leider nicht verstanden'
                    });
                },
                async (step) => {
                    switch (step.result.value) {
                        case cancelChoice.value:
                        break;

                        case deleteChoice.value:
                        await this.eventsService.clearReminder(step.options.id, step.options.rsvp.user.id);
                        step.options.reminder = null;
                        await step.context.sendActivity(MessageFactory.text('Alles klar, für diese Veranstaltung \
                        habe ich die Erinnerung gelöscht.'));
                        break;

                        default:
                        const reminder = new Date(step.result.value);
                        await this.eventsService.setReminder(step.options.id, step.options.rsvp.user.id, reminder);
                        step.options.reminder = reminder.valueOf();
                        await step.context.sendActivity(MessageFactory.text('Alles klar, ich werde dir am '
                        + reminder.toLocaleString('de') + ' eine Erinnerung senden.'));
                    }

                    // prevent any previous user input from getting interpreted as a command
                    step.context.activity.text = '';
                    return await step.endDialog();
                }
            ])
        );

        // On reminder Dialogs
        this.addDialog(
            new WaterfallDialog<EventDialog.EventRecap>(id + EventDialog.EventDialogs.ON_REMINDER_DIALOG, [
                async (step) => {
                    // TODO: display reminder and new reminder setters
                    return await step.next();
                },
                async (step) => {
                    // TODO: jump to the set reminder dialog
                    return await step.next();
                }
            ])
        );

        this.addDialog(new TextPrompt(id + 'tp'));
        this.addDialog(new ChoicePrompt(id + 'cp'));
    }

    public async onBeginDialog(innerDC: DialogContext, options: EventDialog.EventDialogOptions) {
        if (!(options[0] as Event)?.dateStart) {
            const id = 'string' === typeof options[0] ? options[0] : options[0].id;
            options[0] = await this.eventsService.getEvent(id);
        }

        const user = await this.userSettings.get(innerDC.context);

        const event = options[0] as EventDialog.EventRecap;
        event.dateStart = event.dateStart * 1000; // db saves up to second, js uses ms
        event.rsvp = await this.eventsService.getAttendance(event.id, user.privateId);

        if (!event.rsvp?.user) {
            (event.rsvp as DeepPartial<EventAttendee>) = { user: { id: user.privateId } };
        }

        switch (String(options?.length)) {
            case '1':
            await this.sendEventCard(innerDC, event);
            return await innerDC.endDialog();

            case '2':
            return await innerDC.beginDialog(this.id + options[1], event);

            default:
            throw new Error('An incorrect amount of parameters were supplied to this dialog');
        }
    }

    private async sendEventCard(innerDC: DialogContext, recap: EventDialog.EventRecap) {
        let text = this.configService.get('HOME_URL') + '/events/' + recap.id + '\n\n'
            + 'Datum: ' + new Date(recap.dateStart).toLocaleString('de') + '\n\n';

        if (recap.location?.trim().length) {
            text += 'Ort: ' + recap.location + '\n\n';
        }

        if (recap.broadcast) {
            text = recap.broadcast + '\n\n' + text;
        }

        const statusString: EventDialog.ParticipationStrings = {
            accepted: '✅ Angemeldet ✅',
            maybe: '🤔 Mit Vorbehalt 🤔',
            declined: '❌ Abgemeldet ❌'
        };

        await innerDC.context.sendActivity({ attachments: [
            CardFactory.heroCard(
                recap.title,
                CardFactory.images([recap?.imageUrl]),
                CardFactory.actions([
                    {
                        type: 'postBack',
                        title: recap.rsvp ? statusString[recap.rsvp.status] : 'Teilnahme anpassen',
                        value: [EventDialog.name, recap.id, EventDialog.EventDialogs.RSVP_DIALOG].join(',')
                    },
                    {
                        type: 'postBack',
                        title: recap.reminder ? '⏰ ' + new Date(recap.reminder).toLocaleString('de') + ' ⏰' : 'Reminder setzen',
                        value: [EventDialog.name, recap.id, EventDialog.EventDialogs.SET_REMINDER_DIALOG].join(',')
                    }
                ]),
                { text }
            )
        ] });

        // prevent any previous user input from getting interpreted as a command
        innerDC.context.activity.text = '';
    }

    private isInDateRange(eventDate: number, minBound: number) {
        return Date.now() < (eventDate - minBound * 60 * 1000);
    }
}

export namespace EventDialog {
    export type EventDialogOptions = [
        Pick<Event, 'id'>,
        EventDialogs?
    ];

    export interface EventRecap extends Event {
        reminder?: number;
        rsvp?: EventAttendee;
        imgUrl?: string;
        broadcast?: string;
    }

    export enum EventDialogs {
        SET_REMINDER_DIALOG = 'remind',
        ON_REMINDER_DIALOG = 'onReminder',
        RSVP_DIALOG = 'rsvp'
    }

    export type ParticipationStrings = { [K in EventAttendance]: string };
}
