import { Injectable, OnApplicationShutdown } from '@nestjs/common';
import { InjectRepository } from '@nestjs/typeorm';
import { EventEmitter } from 'events';
import { FindConditions, In, IsNull, Raw, Repository } from 'typeorm';

import { Job } from './job.entity';

@Injectable()
export class JobsService extends EventEmitter implements OnApplicationShutdown {

    private timers: NodeJS.Timeout[] = [];

    constructor(
        @InjectRepository(Job)
        private readonly repository: Repository<Job>) {
        super();

        // start the regular job fetching
        this.timers.push(setInterval(() => {
            this.fetchNewJobs();
        }, 60 * 1000));
        this.fetchNewJobs(); // run now
    }

    /**
     * Fetch jobs which are supposed to run in the span of the next 2 minutes and reserve them
     * so no other backend instance processes them as well.
     */
    public fetchNewJobs(): void {
        this.repository.find({ where: {
            status: IsNull(),
            runOn: Raw((alias) => `${alias} < (NOW() + INTERVAL 2 MINUTE)`)
        },
        select: [
            'id', 'runOn'
        ] }).then(async (entities) => {
            if (!entities.length) { return; }
            // mark fetched jobs as WIP
            await this.repository.update({ id: In(entities.map((entity) => entity.id)) }, { status: 'wip' });

            for (const entity of entities) {
                const timeDiffMs = Math.max(entity.runOn.valueOf() - Date.now(), 0);
                this.timers.push(setTimeout(async () => {
                    await this.run(entity);
                }, timeDiffMs));
            }
        }).catch((err) => {
            console.error(err);
        });
    }

    public async add(trigger: string, runOn: Date, args?: any[]) {
        return await this.repository.save({
            trigger,
            runOn,
            args: JSON.stringify(args || [])
        });
    }

    public async one(conditions: FindConditions<Job>): Promise<Job> {
        return this.repository.findOneOrFail(conditions);
    }

    public async remove(jobs: Job[]): Promise<void> {
        await this.repository.remove(jobs);
    }

    /**
     * Trigger the listeners of the given job.
     * @param entity
     */
    private async run(entity: Job) {
        // re-fetch entity in case the job has been modified or deleted while waiting
        entity = await this.one({ id: entity.id });
        super.emit(entity.trigger, JSON.parse(entity.args));
        await this.repository.remove(entity);
    }

    // make this method for external use obsolete
    public emit(event: string | symbol, ...args: any[]): boolean {
        return false;
    }

    /**
     * Clear timers and un-reserve already fetched pending jobs.
     */
    public async onApplicationShutdown(signal?: string) {
        for (const timer of this.timers) {
            timer.unref();
        }
        await this.repository.update({ status: 'wip' }, { status: null });
    }
}
