import * as faker from 'faker';

import { SessionService } from '../src/user/session.service';
import { User } from '../src/user/user.entity';
import { UserService } from '../src/user/user.service';

import { Globals } from './common';

export function userTests() {
    const testGlobals: Globals = this;
    let testUser: User;

    it('Fail creating user without session', async () => {
        const syserr = console.error;
        console.error = jest.fn(); // hide error

        await testGlobals.req
        .put('/api/session')
        .send({ sessionId: faker.internet.password() })
        .expect(500);

        expect(console.error).toHaveBeenCalled();
        console.error = syserr; // restore
    });

    it('Create user', async () => {
        const userService = testGlobals.app.get(UserService);
        const userDto = {
            name: faker.name.firstName(),
            telegramId: faker.random.uuid()
        };

        // bot calls this on the first connection
        const user = await userService.create(userDto.name, userDto.telegramId);

        expect(user).toEqual({
            id: expect.any(String),
            name: userDto.name,
            telegramId: userDto.telegramId,
            publicId: expect.any(String),
            dateCreated: expect.any(Date)
        } as User);

        // save that user for the next tests
        testUser = user;
    });

    it('Get the full user', async () => {
        const { body } = await testGlobals.req
        .get('/api/user/' + testUser.publicId)
        .expect(200);

        const { telegramId, id, dateCreated, ...user } = testUser;

        expect(body).toMatchObject(user);
    });

    it('Create a session', async () => {
        const sessionService = testGlobals.app.get(SessionService);

        // bot sends the user to create a new session
        const session = await sessionService.create(testUser);

        /**
         * Perform session validity tests.
         */
        const testSess = { ...session };
        const MINUTE = 60 * 1000;
        const HOUR = 60 * MINUTE;
        const YEAR = 24 * 365 * HOUR;

        // session is older than an hour, contains user agent = valid
        testSess.dateStarted = Date.now() - (HOUR + MINUTE);
        testSess.userAgent = faker.internet.userAgent();
        expect(sessionService.isValid(testSess)).toBeTruthy();

        // session is older than a year, still contains user agent = invalid
        testSess.dateStarted = Date.now() - (YEAR + MINUTE);
        expect(sessionService.isValid(testSess)).toBeFalsy();

        // remove the client's user agent, still older than a year = invalid
        testSess.userAgent = '';
        expect(sessionService.isValid(testSess)).toBeFalsy();

        // no user agent, session is older than an hour = invalid
        testSess.dateStarted = Date.now() - (HOUR + MINUTE);
        expect(sessionService.isValid(testSess)).toBeFalsy();

        // no user agent, session is younger than an hour = valid
        testSess.dateStarted = Date.now() - (HOUR - MINUTE);
        expect(sessionService.isValid(testSess)).toBeTruthy();

        /**
         * Return to the normal test sequence.
         */

        // the frontend sends the expected session ID
        const res = await testGlobals.req
        .put('/api/session')
        .set('user-agent', faker.internet.userAgent())
        .send({ sessionId: session.id })
        .expect(200);

        testGlobals.cookies = res.get('Set-Cookie');

        expect(testGlobals.cookies.length).toEqual(2);
    });
}

export function userTestsDestroy() {
    const testGlobals: Globals = this;

    it('Delete the user', async () => {
        const { body: user } = await testGlobals.req
        .get('/api/user')
        .set('Cookie', testGlobals.cookies)
        .expect(200);

        await testGlobals.req
        .delete('/api/user')
        .set('Cookie', testGlobals.cookies)
        .expect(200);

        const userService = testGlobals.app.get(UserService);
        const userExists = await userService.exists({ publicId: user.publicId });
        expect(userExists).toBeFalsy();
    });
}
